from typing import Generic

from ground import hints
from ground.enums import Orientation
from typing_extensions import Self

from .event import Event
from .hints import Map, Orienteer, ScalarT


class SweepLineKey(Generic[ScalarT]):
    endpoints: Map[Event, hints.Point[ScalarT]]
    opposites: Map[Event, Event]
    event: Event
    _orienteer: Orienteer[ScalarT]

    __slots__ = '_orienteer', 'endpoints', 'event', 'opposites'

    def __new__(
        cls,
        endpoints: Map[Event, hints.Point[ScalarT]],
        opposites: Map[Event, Event],
        event: Event,
        orienteer: Orienteer[ScalarT],
        /,
    ) -> Self:
        self = super().__new__(cls)
        self.endpoints, self.event, self.opposites, self._orienteer = (
            endpoints,
            event,
            opposites,
            orienteer,
        )
        return self

    def __lt__(self, other: Self, /) -> bool:
        """
        Checks if the segment (or at least the point) associated with event
        is lower than other's.
        """
        assert self.endpoints is other.endpoints
        assert self.opposites is other.opposites
        event, other_event = self.event, other.event
        start, other_start = self.endpoints[event], self.endpoints[other_event]
        end, other_end = (
            self.endpoints[self.opposites[event]],
            self.endpoints[self.opposites[other_event]],
        )
        other_start_orientation = self._orienteer(start, end, other_start)
        other_end_orientation = self._orienteer(start, end, other_end)
        if other_start_orientation is other_end_orientation:
            start_x, start_y = start.x, start.y
            other_start_x, other_start_y = other_start.x, other_start.y
            if other_start_orientation is not Orientation.COLLINEAR:
                # other segment fully lies on one side
                return other_start_orientation is Orientation.COUNTERCLOCKWISE
            # segments are collinear
            if start_y == other_start_y:
                end_x, end_y = end.x, end.y
                other_end_x, other_end_y = other_end.x, other_end.y
                if start_x != other_start_x:
                    return start_x < other_start_x
                # segments have same start
                if end_y != other_end_y:
                    return end_y < other_end_y
                # segments are horizontal
                return end_x < other_end_x
            return start_y < other_start_y
        start_orientation = self._orienteer(other_start, other_end, start)
        end_orientation = self._orienteer(other_start, other_end, end)
        if start_orientation is end_orientation:
            return start_orientation is Orientation.CLOCKWISE
        if other_start_orientation is Orientation.COLLINEAR:
            return other_end_orientation is Orientation.COUNTERCLOCKWISE
        if start_orientation is Orientation.COLLINEAR:
            return end_orientation is Orientation.CLOCKWISE
        if end_orientation is Orientation.COLLINEAR:
            return start_orientation is Orientation.CLOCKWISE
        return other_start_orientation is Orientation.COUNTERCLOCKWISE



from __future__ import annotations

from typing import Generic, TYPE_CHECKING

from ground.hints import Point
from typing_extensions import Self

from .event import Event, is_event_left
from .hints import ScalarT

if TYPE_CHECKING:
    from .hints import Map


class EventsQueueKey(Generic[ScalarT]):
    endpoints: Map[Event, Point[ScalarT]]
    opposites: Map[Event, Event]
    event: Event

    __slots__ = 'endpoints', 'event', 'opposites'

    def __new__(
        cls,
        endpoints: Map[Event, Point[ScalarT]],
        opposites: Map[Event, Event],
        event: Event,
        /,
    ) -> Self:
        self = super().__new__(cls)
        self.endpoints, self.event, self.opposites = (
            endpoints,
            event,
            opposites,
        )
        return self

    def __lt__(self, other: Self) -> bool:
        """
        Checks if the event should be processed before the other.
        """
        event, other_event = self.event, other.event
        event_start, other_event_start = (
            self.endpoints[event],
            self.endpoints[other_event],
        )
        start_x, start_y = event_start.x, event_start.y
        other_start_x, other_start_y = other_event_start.x, other_event_start.y
        if start_x != other_start_x:
            # different x-coordinate,
            # the event with lower x-coordinate is processed first
            return start_x < other_start_x
        if start_y != other_start_y:
            # different starts, but same x-coordinate,
            # the event with lower y-coordinate is processed first
            return start_y < other_start_y
        if is_event_left(event) is not is_event_left(other_event):
            # same start, but one is a left endpoint
            # and the other is a right endpoint,
            # the right endpoint is processed first
            return not is_event_left(event)
        # same start,
        # both events are left endpoints or both are right endpoints
        return (
            self.endpoints[self.opposites[event]]
            < self.endpoints[self.opposites[other_event]]
        )



from typing import NewType

from .utils import is_even

Event = NewType('Event', int)

is_event_left = is_even


def segment_id_to_left_event(segment_id: int, /) -> Event:
    return Event(2 * segment_id)


def segment_id_to_right_event(segment_id: int, /) -> Event:
    return Event(2 * segment_id + 1)



from collections.abc import Sequence as _Sequence

from ground.context import Context as _Context
from ground.enums import Relation as _Relation
from ground.hints import Contour as _Contour, Segment as _Segment

from ._core.base import sweep as _sweep
from ._core.hints import ScalarT as _ScalarT
from ._core.utils import all_unique as _all_unique


def contour_self_intersects(
    contour: _Contour[_ScalarT], /, *, context: _Context[_ScalarT]
) -> bool:
    """
    Checks if contour has self-intersection.

    Based on Bentley-Ottmann algorithm.

    Time complexity:
        ``O(len(contour.vertices) * log len(contour.vertices))``
    Memory complexity:
        ``O(len(contour.vertices))``
    Reference:
        https://en.wikipedia.org/wiki/Sweep_line_algorithm

    :param contour: contour to check.
    :param context: geometrical context.
    :returns: true if contour is self-intersecting, false otherwise.

    .. note::
        Consecutive equal vertices like ``Point(2, 0)`` in

        .. code-block:: python

            Contour([Point(0, 0), Point(2, 0), Point(2, 0), Point(2, 2)])

        will be considered as self-intersection,
        if you don't want them to be treated as such
        -- filter out before passing as argument.

    >>> import math
    >>> from fractions import Fraction
    >>> from ground.context import Context
    >>> context = Context(coordinate_factory=Fraction, sqrt=math.sqrt)
    >>> Contour, Point = context.contour_cls, context.point_cls
    >>> contour_self_intersects(
    ...     Contour([Point(0, 0), Point(2, 0), Point(2, 2)]), context=context
    ... )
    False
    >>> contour_self_intersects(
    ...     Contour([Point(0, 0), Point(2, 0), Point(1, 0)]), context=context
    ... )
    True
    """
    vertices = contour.vertices
    if len(vertices) < 3:
        raise ValueError(f'Contour {contour} is degenerate.')
    if not _all_unique(vertices):
        return True
    segments = context.contour_segments(contour)
    return not all(
        event.relation in (_Relation.DISJOINT, _Relation.TOUCH)
        for event in _sweep(
            segments,
            context.angle_orientation,
            lambda first_start, first_end, second_start, second_end: (
                context.segments_intersection(
                    context.segment_cls(first_start, first_end),
                    context.segment_cls(second_start, second_end),
                )
            ),
        )
    )


def segments_intersect(
    segments: _Sequence[_Segment[_ScalarT]], /, *, context: _Context[_ScalarT]
) -> bool:
    """
    Checks if segments have at least one intersection.

    Based on Shamos-Hoey algorithm.

    Time complexity:
        ``O(len(segments) * log len(segments))``
    Memory complexity:
        ``O(len(segments))``
    Reference:
        https://en.wikipedia.org/wiki/Sweep_line_algorithm

    :param segments: sequence of segments.
    :param context: geometrical context.
    :returns: true if segments intersection found, false otherwise.

    >>> import math
    >>> from fractions import Fraction
    >>> from ground.context import Context
    >>> context = Context(coordinate_factory=Fraction, sqrt=math.sqrt)
    >>> Point, Segment = context.point_cls, context.segment_cls
    >>> segments_intersect([], context=context)
    False
    >>> segments_intersect(
    ...     [Segment(Point(0, 0), Point(2, 2))], context=context
    ... )
    False
    >>> segments_intersect(
    ...     [
    ...         Segment(Point(0, 0), Point(2, 0)),
    ...         Segment(Point(0, 2), Point(2, 2)),
    ...     ],
    ...     context=context,
    ... )
    False
    >>> segments_intersect(
    ...     [
    ...         Segment(Point(0, 0), Point(2, 2)),
    ...         Segment(Point(0, 0), Point(2, 2)),
    ...     ],
    ...     context=context,
    ... )
    True
    >>> segments_intersect(
    ...     [
    ...         Segment(Point(0, 0), Point(2, 2)),
    ...         Segment(Point(2, 0), Point(0, 2)),
    ...     ],
    ...     context=context,
    ... )
    True
    """
    return not all(
        event.relation is _Relation.DISJOINT
        for event in _sweep(
            segments,
            context.angle_orientation,
            lambda first_start, first_end, second_start, second_end: (
                context.segments_intersection(
                    context.segment_cls(first_start, first_end),
                    context.segment_cls(second_start, second_end),
                )
            ),
        )
    )


def segments_cross_or_overlap(
    segments: _Sequence[_Segment[_ScalarT]], /, *, context: _Context[_ScalarT]
) -> bool:
    """
    Checks if at least one pair of segments crosses or overlaps.

    Based on Bentley-Ottmann algorithm.

    Time complexity:
        ``O(len(segments) * log len(segments))``
    Memory complexity:
        ``O(len(segments))``
    Reference:
        https://en.wikipedia.org/wiki/Bentley%E2%80%93Ottmann_algorithm

    :param segments: sequence of segments.
    :param context: geometrical context.
    :returns: true if segments overlap or cross found, false otherwise.

    >>> import math
    >>> from fractions import Fraction
    >>> from ground.context import Context
    >>> context = Context(coordinate_factory=Fraction, sqrt=math.sqrt)
    >>> Point, Segment = context.point_cls, context.segment_cls
    >>> segments_cross_or_overlap([], context=context)
    False
    >>> segments_cross_or_overlap(
    ...     [Segment(Point(0, 0), Point(2, 2))], context=context
    ... )
    False
    >>> segments_cross_or_overlap(
    ...     [
    ...         Segment(Point(0, 0), Point(2, 0)),
    ...         Segment(Point(0, 2), Point(2, 2)),
    ...     ],
    ...     context=context,
    ... )
    False
    >>> segments_cross_or_overlap(
    ...     [
    ...         Segment(Point(0, 0), Point(2, 2)),
    ...         Segment(Point(0, 0), Point(2, 2)),
    ...     ],
    ...     context=context,
    ... )
    True
    >>> segments_cross_or_overlap(
    ...     [
    ...         Segment(Point(0, 0), Point(2, 2)),
    ...         Segment(Point(0, 2), Point(2, 0)),
    ...     ],
    ...     context=context,
    ... )
    True
    """
    return not all(
        event.relation in (_Relation.DISJOINT, _Relation.TOUCH)
        for event in _sweep(
            segments,
            context.angle_orientation,
            lambda first_start, first_end, second_start, second_end: (
                context.segments_intersection(
                    context.segment_cls(first_start, first_end),
                    context.segment_cls(second_start, second_end),
                )
            ),
        )
    )



from collections.abc import Hashable, Iterable
from typing import Protocol, TypeVar

from typing_extensions import Self


class Ordered(Protocol):
    def __lt__(self, other: Self, /) -> bool: ...


_HashableT = TypeVar('_HashableT', bound=Hashable)


def all_unique(values: Iterable[_HashableT]) -> bool:
    seen: set[_HashableT] = set()
    seen_add = seen.add
    for value in values:
        if value in seen:
            return False
        seen_add(value)
    return True


def is_even(value: int, /) -> bool:
    return not (value & 1)


_OrderedT = TypeVar('_OrderedT', bound=Ordered)


def to_sorted_pair(
    first: _OrderedT, second: _OrderedT, /
) -> tuple[_OrderedT, _OrderedT]:
    return (first, second) if first < second else (second, first)



from collections.abc import Iterable, Sequence
from itertools import combinations
from typing import Generic

from ground import hints
from ground.enums import Relation
from typing_extensions import Self

from .events_registry import EventsRegistry
from .hints import Orienteer, ScalarT, SegmentsIntersector


class Intersection(Generic[ScalarT]):
    @property
    def end(self, /) -> hints.Point[ScalarT]:
        return self._end

    @property
    def first_segment_id(self, /) -> int:
        return self._first_segment_id

    @property
    def relation(self, /) -> Relation:
        return self._relation

    @property
    def second_segment_id(self, /) -> int:
        return self._second_segment_id

    @property
    def start(self, /) -> hints.Point[ScalarT]:
        return self._start

    _end: hints.Point[ScalarT]
    _first_segment_id: int
    _relation: Relation
    _second_segment_id: int
    _start: hints.Point[ScalarT]

    __slots__ = (
        '_end',
        '_first_segment_id',
        '_relation',
        '_second_segment_id',
        '_start',
    )

    def __new__(
        cls,
        first_segment_id: int,
        second_segment_id: int,
        relation: Relation,
        start: hints.Point[ScalarT],
        end: hints.Point[ScalarT],
        /,
    ) -> Self:
        self = super().__new__(cls)
        (
            self._end,
            self._first_segment_id,
            self._relation,
            self._second_segment_id,
            self._start,
        ) = (end, first_segment_id, relation, second_segment_id, start)
        return self


def sweep(
    segments: Sequence[hints.Segment[ScalarT]],
    orienteer: Orienteer[ScalarT],
    segments_intersector: SegmentsIntersector[ScalarT],
    /,
) -> Iterable[Intersection[ScalarT]]:
    events_registry = EventsRegistry.from_segments(
        segments, orienteer, segments_intersector, unique=False
    )
    events = iter(events_registry)
    try:
        event = next(events)
    except StopIteration:
        return
    start = events_registry.to_event_start(event)
    segments_ids_containing_start = [
        events_registry.to_event_segment_id(event)
    ]
    for event in events:
        event_start = events_registry.to_event_start(event)
        if event_start == start:
            segments_ids_containing_start.append(
                events_registry.to_event_segment_id(event)
            )
        else:
            yield from segments_ids_containing_point_to_intersections(
                segments_ids_containing_start, start, events_registry
            )
            start = event_start
            segments_ids_containing_start = [
                events_registry.to_event_segment_id(event)
            ]
    yield from segments_ids_containing_point_to_intersections(
        segments_ids_containing_start, start, events_registry
    )


def segments_ids_containing_point_to_intersections(
    segments_ids: Sequence[int],
    point: hints.Point[ScalarT],
    events_registry: EventsRegistry[ScalarT],
    /,
) -> Iterable[Intersection[ScalarT]]:
    for first_segment_id, second_segment_id in combinations(segments_ids, 2):
        first_start = events_registry.to_segment_start(first_segment_id)
        first_end = events_registry.to_segment_end(first_segment_id)
        second_start = events_registry.to_segment_start(second_segment_id)
        second_end = events_registry.to_segment_end(second_segment_id)
        if first_segment_id == second_segment_id:
            start, end = first_start, first_end
            relation = Relation.EQUAL
        elif not events_registry.are_collinear(
            first_segment_id, second_segment_id
        ):
            if (
                first_start == point
                or first_end == point
                or second_start == point
                or second_end == point
            ):
                relation = Relation.TOUCH
            else:
                relation = Relation.CROSS
            start = end = point
        elif max(first_start, second_start) == min(first_end, second_end):
            relation = Relation.TOUCH
            start = end = point
        elif first_start == second_start:
            start = first_start
            if first_end == second_end:
                relation = Relation.EQUAL
                end = first_end
            elif first_end > second_end:
                relation = Relation.COMPOSITE
                end = second_end
            else:
                relation = Relation.COMPONENT
                end = first_end
        elif first_start > second_start:
            start = first_start
            if first_end > second_end:
                relation = Relation.OVERLAP
                end = second_end
            else:
                relation = Relation.COMPONENT
                end = first_end
        elif first_end < second_end:
            relation = Relation.OVERLAP
            start, end = second_start, first_end
        else:
            relation = Relation.COMPOSITE
            start, end = second_start, second_end
        yield Intersection(
            first_segment_id, second_segment_id, relation, start, end
        )



from __future__ import annotations

from typing import Generic, TYPE_CHECKING

from dendroid import red_black
from ground import hints
from ground.enums import Orientation
from prioq.base import PriorityQueue

from .event import (
    Event,
    is_event_left,
    segment_id_to_left_event,
    segment_id_to_right_event,
)
from .events_queue_key import EventsQueueKey
from .hints import ScalarT
from .sweep_line_key import SweepLineKey
from .utils import to_sorted_pair

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from dendroid.hints import KeyedSet

    from .hints import Orienteer, SegmentsIntersector


class EventsRegistry(Generic[ScalarT]):
    @classmethod
    def from_segments(
        cls,
        segments: Sequence[hints.Segment[ScalarT]],
        orienteer: Orienteer[ScalarT],
        segments_intersector: SegmentsIntersector[ScalarT],
        /,
        *,
        unique: bool,
    ) -> EventsRegistry[ScalarT]:
        result = cls(orienteer, segments_intersector, unique)
        for segment_id, segment in enumerate(segments):
            left_event = segment_id_to_left_event(segment_id)
            right_event = segment_id_to_right_event(segment_id)
            start, end = to_sorted_pair(segment.start, segment.end)
            if start == end:
                raise ValueError(
                    'Degenerate segment found '
                    f'with both endpoints being: {start}.'
                )
            result._endpoints.append(start)
            result._endpoints.append(end)
            result._opposites.append(right_event)
            result._opposites.append(left_event)
            result._push(left_event)
            result._push(right_event)
        result._min_collinear_segments_ids = list(range(len(segments)))
        result._segments_ids = list(range(len(segments)))
        return result

    @property
    def unique(self, /) -> bool:
        return self._unique

    def are_collinear(
        self, first_segment_id: int, second_segment_id: int, /
    ) -> bool:
        return self._to_min_collinear_segment_id(
            first_segment_id
        ) == self._to_min_collinear_segment_id(second_segment_id)

    def to_event_end(self, event: Event, /) -> hints.Point[ScalarT]:
        return self.to_event_start(self._to_opposite_event(event))

    def to_event_segment_id(self, event: Event, /) -> int:
        return self._to_left_event_segment_id(
            event if is_event_left(event) else self._to_opposite_event(event)
        )

    def to_event_start(self, event: Event, /) -> hints.Point[ScalarT]:
        return self._endpoints[event]

    def to_segment_end(self, segment_id: int, /) -> hints.Point[ScalarT]:
        return self.to_event_start(segment_id_to_right_event(segment_id))

    def to_segment_start(self, segment_id: int, /) -> hints.Point[ScalarT]:
        return self.to_event_start(segment_id_to_left_event(segment_id))

    __slots__ = (
        '_endpoints',
        '_events_queue_data',
        '_min_collinear_segments_ids',
        '_opposites',
        '_orienteer',
        '_segments_ids',
        '_segments_intersector',
        '_sweep_line_data',
        '_unique',
    )

    def __init__(
        self,
        orienteer: Orienteer[ScalarT],
        segments_intersector: SegmentsIntersector[ScalarT],
        unique: bool,  # noqa: FBT001
        /,
    ) -> None:
        self._orienteer, self._segments_intersector, self._unique = (
            orienteer,
            segments_intersector,
            unique,
        )
        self._opposites: list[Event] = []
        self._endpoints: list[hints.Point[ScalarT]] = []
        self._segments_ids: list[int] = []
        self._min_collinear_segments_ids: list[int] = []
        self._events_queue_data: PriorityQueue[
            EventsQueueKey[ScalarT], Event
        ] = PriorityQueue(
            key=lambda event: EventsQueueKey(
                self._endpoints, self._opposites, event
            )
        )
        self._sweep_line_data: KeyedSet[SweepLineKey[ScalarT], Event] = (
            red_black.set_(key=self._event_to_sweep_line_key)
        )

    def _event_to_sweep_line_key(
        self, event: Event, /
    ) -> SweepLineKey[ScalarT]:
        return SweepLineKey(
            self._endpoints, self._opposites, event, self._orienteer
        )

    def __bool__(self) -> bool:
        return bool(self._events_queue_data)

    def __iter__(self) -> Iterator[Event]:
        while self:
            event = self._pop()
            if is_event_left(event):
                equal_segment_event = self._find(event)
                if equal_segment_event is None:
                    self._add(event)
                    below_event = self._below(event)
                    if below_event is not None:
                        self._detect_intersection(below_event, event)
                    above_event = self._above(event)
                    if above_event is not None:
                        self._detect_intersection(event, above_event)
                    yield event
                else:
                    self._merge_equal_segment_events(
                        equal_segment_event, event
                    )
                    if not self.unique:
                        yield event
            else:
                event_opposite = self._to_opposite_event(event)
                equal_segment_event = self._find(event_opposite)
                if equal_segment_event is not None:
                    above_event, below_event = (
                        self._above(equal_segment_event),
                        self._below(equal_segment_event),
                    )
                    self._remove(equal_segment_event)
                    if below_event is not None and above_event is not None:
                        self._detect_intersection(below_event, above_event)
                    if event != equal_segment_event:
                        self._merge_equal_segment_events(
                            equal_segment_event, event_opposite
                        )
                    yield event
                elif not self.unique:
                    yield event

    def _above(self, event: Event, /) -> Event | None:
        assert is_event_left(event)
        try:
            return self._sweep_line_data.next(event)
        except ValueError:
            return None

    def _add(self, event: Event, /) -> None:
        assert is_event_left(event)
        self._sweep_line_data.add(event)

    def _below(self, event: Event, /) -> Event | None:
        assert is_event_left(event)
        try:
            return self._sweep_line_data.prev(event)
        except ValueError:
            return None

    def _detect_intersection(
        self, below_event: Event, event: Event, /
    ) -> None:
        event_start = self.to_event_start(event)
        event_end = self.to_event_end(event)
        below_event_start = self.to_event_start(below_event)
        below_event_end = self.to_event_end(below_event)
        event_start_orientation = self._orienteer(
            below_event_end, below_event_start, event_start
        )
        event_end_orientation = self._orienteer(
            below_event_end, below_event_start, event_end
        )
        if event_start_orientation is event_end_orientation:
            if event_start_orientation is Orientation.COLLINEAR:
                if event_start == below_event_start:
                    assert event_end != below_event_end, (
                        event_start,
                        event_end,
                        below_event_start,
                        below_event_end,
                    )
                    max_end_event, min_end_event = (
                        (below_event, event)
                        if event_end < below_event_end
                        else (event, below_event)
                    )
                    self._remove(max_end_event)
                    min_end = self.to_event_end(min_end_event)
                    _, min_end_max_end_event = self._divide(
                        max_end_event, min_end
                    )
                    self._push(min_end_max_end_event)
                    self._merge_equal_segment_events(event, below_event)
                elif event_end == below_event_end:
                    max_start_event, min_start_event = (
                        (below_event, event)
                        if event_start < below_event_start
                        else (event, below_event)
                    )
                    max_start = self.to_event_start(max_start_event)
                    (max_start_to_min_start_event, max_start_to_end_event) = (
                        self._divide(min_start_event, max_start)
                    )
                    self._push(max_start_to_min_start_event)
                    self._merge_equal_segment_events(
                        max_start_event, max_start_to_end_event
                    )
                elif below_event_start < event_start < below_event_end:
                    if event_end < below_event_end:
                        self._divide_event_by_mid_segment_event_endpoints(
                            below_event, event, event_start, event_end
                        )
                    else:
                        max_start, min_end = event_start, below_event_end
                        self._divide_overlapping_events(
                            below_event, event, max_start, min_end
                        )
                elif event_start < below_event_start < event_end:
                    if below_event_end < event_end:
                        self._divide_event_by_mid_segment_event_endpoints(
                            event,
                            below_event,
                            below_event_start,
                            below_event_end,
                        )
                    else:
                        max_start, min_end = below_event_start, event_end
                        self._divide_overlapping_events(
                            event, below_event, max_start, min_end
                        )
        elif event_start_orientation is Orientation.COLLINEAR:
            if below_event_start < event_start < below_event_end:
                point = event_start
                self._divide_event_by_midpoint(below_event, point)
        elif event_end_orientation is Orientation.COLLINEAR:
            if below_event_start < event_end < below_event_end:
                point = event_end
                self._divide_event_by_midpoint(below_event, point)
        else:
            below_event_start_orientation = self._orienteer(
                event_start, event_end, below_event_start
            )
            below_event_end_orientation = self._orienteer(
                event_start, event_end, below_event_end
            )
            if below_event_start_orientation is Orientation.COLLINEAR:
                assert below_event_end_orientation is not Orientation.COLLINEAR
                if event_start < below_event_start < event_end:
                    point = below_event_start
                    self._divide_event_by_midpoint_checking_above(event, point)
            elif below_event_end_orientation is Orientation.COLLINEAR:
                if event_start < below_event_end < event_end:
                    point = below_event_end
                    self._divide_event_by_midpoint_checking_above(event, point)
            elif (
                below_event_start_orientation
                is not below_event_end_orientation
            ):
                cross_point = self._segments_intersector(
                    event_start, event_end, below_event_start, below_event_end
                )
                if below_event_start < cross_point < below_event_end:
                    self._divide_event_by_midpoint(below_event, cross_point)
                if event_start < cross_point < event_end:
                    self._divide_event_by_midpoint_checking_above(
                        event, cross_point
                    )

    def _divide(
        self, event: Event, mid_point: hints.Point[ScalarT], /
    ) -> tuple[Event, Event]:
        assert is_event_left(event)
        opposite_event = self._to_opposite_event(event)
        mid_point_to_event_end_event = Event(len(self._endpoints))
        self._segments_ids.append(self._to_left_event_segment_id(event))
        self._endpoints.append(mid_point)
        self._opposites.append(opposite_event)
        self._opposites[opposite_event] = mid_point_to_event_end_event
        mid_point_to_event_start_event = Event(len(self._endpoints))
        self._endpoints.append(mid_point)
        self._opposites.append(event)
        self._opposites[event] = mid_point_to_event_start_event
        return mid_point_to_event_start_event, mid_point_to_event_end_event

    def _divide_event_by_mid_segment_event_endpoints(
        self,
        event: Event,
        mid_segment_event: Event,
        mid_segment_event_start: hints.Point[ScalarT],
        mid_segment_event_end: hints.Point[ScalarT],
        /,
    ) -> None:
        self._divide_event_by_midpoint(event, mid_segment_event_end)
        (
            mid_segment_event_start_to_event_start_event,
            min_segment_event_start_to_min_segment_event_end_event,
        ) = self._divide(event, mid_segment_event_start)
        self._push(mid_segment_event_start_to_event_start_event)
        self._merge_equal_segment_events(
            mid_segment_event,
            min_segment_event_start_to_min_segment_event_end_event,
        )

    def _divide_event_by_midpoint(
        self, event: Event, point: hints.Point[ScalarT], /
    ) -> None:
        point_to_event_start_event, point_to_event_end_event = self._divide(
            event, point
        )
        self._push(point_to_event_start_event)
        self._push(point_to_event_end_event)

    def _divide_event_by_midpoint_checking_above(
        self, event: Event, point: hints.Point[ScalarT], /
    ) -> None:
        above_event = self._above(event)
        if above_event is not None and (
            self.to_event_start(above_event) == self.to_event_start(event)
            and self.to_event_end(above_event) == point
        ):
            self._remove(above_event)
            self._divide_event_by_midpoint(event, point)
            self._merge_equal_segment_events(event, above_event)
            return
        self._divide_event_by_midpoint(event, point)

    def _divide_overlapping_events(
        self,
        min_start_event: Event,
        max_start_event: Event,
        max_start: hints.Point[ScalarT],
        min_end: hints.Point[ScalarT],
        /,
    ) -> None:
        self._divide_event_by_midpoint(max_start_event, min_end)
        (max_start_to_min_start_event, max_start_to_min_end_event) = (
            self._divide(min_start_event, max_start)
        )
        self._push(max_start_to_min_start_event)
        self._merge_equal_segment_events(
            max_start_event, max_start_to_min_end_event
        )

    def _find(self, event: Event, /) -> Event | None:
        assert is_event_left(event)
        try:
            candidate = self._sweep_line_data.floor(event)
        except ValueError:
            return None
        else:
            return (
                candidate
                if (
                    (
                        self.to_event_start(candidate)
                        == self.to_event_start(event)
                    )
                    and (
                        self.to_event_end(candidate)
                        == self.to_event_end(event)
                    )
                )
                else None
            )

    def _merge_equal_segment_events(
        self, first: Event, second: Event, /
    ) -> None:
        first_segment_id = self._to_left_event_segment_id(first)
        second_segment_id = self._to_left_event_segment_id(second)
        first_min_collinear_segment_id = self._min_collinear_segments_ids[
            first_segment_id
        ]
        second_min_collinear_segment_id = self._min_collinear_segments_ids[
            second_segment_id
        ]
        min_collinear_segment_id = min(
            first_min_collinear_segment_id, second_min_collinear_segment_id
        )
        self._min_collinear_segments_ids[first_segment_id] = (
            min_collinear_segment_id
        )
        self._min_collinear_segments_ids[second_segment_id] = (
            min_collinear_segment_id
        )
        self._min_collinear_segments_ids[first_min_collinear_segment_id] = (
            min_collinear_segment_id
        )
        self._min_collinear_segments_ids[second_min_collinear_segment_id] = (
            min_collinear_segment_id
        )

    def _pop(self) -> Event:
        return self._events_queue_data.pop()

    def _push(self, event: Event, /) -> None:
        self._events_queue_data.push(event)

    def _remove(self, event: Event, /) -> None:
        assert is_event_left(event)
        self._sweep_line_data.remove(event)

    def _to_left_event_segment_id(self, event: Event, /) -> int:
        assert is_event_left(event)
        return self._segments_ids[event // 2]

    def _to_min_collinear_segment_id(self, segment_id: int, /) -> int:
        candidate = segment_id
        iterations_count = 0
        while self._min_collinear_segments_ids[candidate] != candidate:
            candidate = self._min_collinear_segments_ids[candidate]
            iterations_count += 1
        assert iterations_count <= (
            len(self._segments_ids).bit_length() - 1
        ), (iterations_count, self._min_collinear_segments_ids)
        return candidate

    def _to_opposite_event(self, event: Event, /) -> Event:
        return self._opposites[event]



from setuptools import find_packages, setup

project_base_url = 'https://github.com/lycantropos/bentley_ottmann/'


setup(
    packages=find_packages(exclude=('tests', 'tests.*')),
    url=project_base_url,
    download_url=project_base_url + 'archive/master.zip',
)



from __future__ import annotations

from collections.abc import Callable
from typing import TypeVar

from ground import hints
from ground.enums import Orientation
from typing_extensions import Protocol

_Key = TypeVar('_Key', contravariant=True)
_Value = TypeVar('_Value', covariant=True)


class Map(Protocol[_Key, _Value]):
    def __getitem__(self, key: _Key, /) -> _Value: ...


ScalarT = TypeVar('ScalarT', bound=hints.Scalar)
Orienteer = Callable[
    [hints.Point[ScalarT], hints.Point[ScalarT], hints.Point[ScalarT]],
    Orientation,
]
SegmentsIntersector = Callable[
    [
        hints.Point[ScalarT],
        hints.Point[ScalarT],
        hints.Point[ScalarT],
        hints.Point[ScalarT],
    ],
    hints.Point[ScalarT],
]
